@@ -33,7 +33,7 @@ import numpy as np
 import pandas as pd
 
 from qary import constants
-from qary.skills.basebots import normalize_replies
+from qary.skills.base import normalize_replies
 
 __author__ = "see AUTHORS.md and README.md: Travis, Nima, Erturgrul, Aliya, Xavier, Maria, Hobson, ..."
 __copyright__ = "Hobson Lane"
@@ -67,34 +67,36 @@ class CLIBot:
             bots=constants.DEFAULT_BOTS,
             num_top_replies=None,
             **quality_kwargs):
+        if isinstance(bots, str):
+            bots = bots.split(',')
         if not isinstance(bots, abc.Mapping):
-            bots = dict(zip(bots, [None] * len(bots)))
+            bots = dict(zip(bots, [{}] * len(bots)))
+        log.info(f'CLIBot(bots={bots}')
         for bot_name, bot_kwargs in bots.items():
             bot_kwargs = {} if bot_kwargs is None else dict(bot_kwargs)
-            self.add_bot(bot_name, **bot_kwargs)
+            self.add_skill(bot_name, **bot_kwargs)
         self.num_top_replies = constants.DEFAULT_CONFIG['num_top_replies'] if num_top_replies is None else min(
             max(int(num_top_replies), 1), 10000)
         self.repliers = [bot.reply if hasattr(bot, 'reply') else bot for bot in self.bots if bot is not None]
-        log.warning(f'Loaded bots: {self.repliers}')
+        log.debug(f'Loaded skills: {self.bots}')
         self.quality_score = QualityScore(**quality_kwargs)
 
-    def add_bot(self, bot_name, **bot_kwargs):
-        bot_name = bot_name if bot_name.endswith('_bots') else f'{bot_name}_bots'
+    def add_skill(self, bot_name, **bot_kwargs):
         log.info(f'Adding bot named {bot_name}')
         self.bot_names.append(bot_name)
         bot_module = importlib.import_module(f'qary.skills.{bot_name}')
-        new_modules, new_bots = [], []
+        new_modules, new_skills = [], []
         if bot_module.__class__.__name__ == 'module':
             module_vars = tuple(vars(bot_module).items())
-            for bot_class in (c for n, c in module_vars if n.endswith('Bot') and callable(getattr(c, 'reply', None))):
-                log.info(f'Adding a Bot class {bot_class} from module {bot_module}...')
-                new_bots.append(bot_class(**bot_kwargs))
+            for bot_class in (c for n, c in module_vars if n.endswith('Skill') and callable(getattr(c, 'reply', None))):
+                log.info(f'Adding a Skill class {bot_class} from module {bot_module}...')
+                new_skills.append(bot_class(**bot_kwargs))
                 new_modules.append(bot_module)
         else:
             log.warning(f'FIXME: add feature to allow import of specific bot classes like "qary.skills.{bot_name}"')
-        self.bots.extend(new_bots)
+        self.bots.extend(new_skills)
         self.bot_modules.extend(new_modules)
-        return new_bots
+        return new_skills
 
     def log_reply(self, statement, reply):
         history_path = os.path.join(constants.DATA_DIR, 'history.json')
@@ -118,7 +120,7 @@ class CLIBot:
         with open(history_path, 'w') as f:
             json.dump(history, f)
 
-    def reply(self, statement='', context=''):
+    def reply(self, statement='', context=None):
         ''' Collect replies from from loaded bots and return best reply (str). '''
         log.info(f'statement={statement}')
         replies = []
@@ -131,7 +133,7 @@ class CLIBot:
                 bot_replies = bot.reply(statement, context=context)
                 log.info("{bot.__name__} replies: {bot_replies}")
             except TypeError as e:
-                log.warning(f"TypeError: bot.reply probably got an unexpected keyword argument 'context': {e}")
+                log.debug(f"TypeError: bot.reply probably got an unexpected keyword argument 'context': {e}")
                 bot_replies = bot.reply(statement)
             except Exception as e:
                 log.error(f'Error trying to run {bot.__self__.__class__}.reply("{statement}", context={context})')
@@ -143,33 +145,36 @@ class CLIBot:
             replies.extend(bot_replies)
 
         # Weighted random selection of reply from those with top n confidence scores
+        log.info(f"{len(replies)} replies from {len(self.bots)} bots:")
+        log.info(repr(replies))
         if len(replies):
             log.info(f'Found {len(replies)} suitable replies, limiting to {self.num_top_replies}...')
             replies = self.quality_score.update_replies(replies, statement)
             replies = sorted(replies, reverse=True)[:self.num_top_replies]
 
-            confidences, texts = list(zip(*replies))
-            conf_sums = np.cumsum(confidences)
+            conf_sums = np.cumsum(list(r[0] for r in replies))
             roll = np.random.rand() * conf_sums[-1]
 
             for i, threshold in enumerate(conf_sums):
                 if roll < threshold:
-                    reply = texts[i]
-                    self.log_reply(statement, reply)
+                    reply = replies[i][1]
+                    self.log_reply(statement, replies[i][0])
                     return reply
-            log.error(f"Error rolling dice to select reply. roll: {roll}\nthreshold: {threshold}\nconf_sums: {conf_sums}\n"
-                      f"replies: {pd.DataFrame(replies)}")
-            return texts[0]
+            log.error(f"Error rolling dice to select reply."
+                      f"\n    roll:{roll}"
+                      f"\n    threshold: {threshold}"
+                      f"\n    conf_sums: {conf_sums}\n"
+                      f"\n    replies: {pd.DataFrame(replies)}")
+            return replies[0][1]
         else:
             log.error(f"No replies ({replies}) were returned by {self.bots}!!")
 
-        # TODO: np.choice from list of more friendly random unknown error replies...
         reply = "Sorry, something went wrong. Not sure what to say..."
         self.log_reply(statement, reply)
         return reply
 
 
-def run_bot():
+def run_skill():
     global BOT
     if BOT is None:
         BOT = CLIBot(
@@ -185,33 +190,42 @@ def run_bot():
     return BOT
 
 
-def cli(args):
+def cli(args,
+        exit_commands='exit quit bye goodbye cya hasta seeya'.split(),
+        max_turns=constants.MAX_TURNS):
     global BOT
-    BOT = run_bot() if BOT is None else BOT
+    BOT = run_skill() if BOT is None else BOT
     state = {}
-    statements = []
-    user_statement = ' '.join(args.words)
-    statements.append(dict(user=user_statement, bot=None, **state))
-    args.persist = args.persist or not len(user_statement)
-    for i in range(constants.MAX_TURNS if args.persist else 0):
-        if user_statement.lower().strip() in constants.EXIT_COMMANDS:
+    user_statement = ' '.join(args.words).strip() or None
+    if user_statement is not None:
+        max_turns = 0
+    history = [dict(user=user_statement, bot=None, **state)]
+    log.warning(f'user_cli_statement: `{user_statement}`')
+    while True:
+        if user_statement and user_statement.lower().strip() in exit_commands:
             break
-        if user_statement:
-            log.info(f"Computing a reply to {user_statement}...")
-            bot_statement = BOT.reply(user_statement)
-            statements[-1]['bot'] = bot_statement
-            print(f"{args.nickname}: {bot_statement}")
-        if args.persist or not user_statement:
-            user_statement = input("YOU: ")
-            statements.append(dict(user=user_statement, bot=None, **state))
+        log.info(f"Computing a reply to {user_statement}...")
+        if user_statement is None:
+            bot_statement = None
         else:
-            break
-    return pd.DataFrame(statements)
+            bot_statement = BOT.reply(user_statement)
+        history[-1]['bot'] = bot_statement
+        if bot_statement is None:
+            pass
+            # print()
+            # print(f"{args.nickname}: Hi! I'm **{args.nickname}**. Ask me anything!")
+        else:
+            print(f'{args.nickname}: ')
+            print(bot_statement)
+        if max_turns > 0 or not user_statement:
+            user_statement = input("YOU: ")
+            history.append(dict(user=user_statement, skill=None, **state))
+    return pd.DataFrame(history)
 
 
 def main():
     global BOT
-    BOT = run_bot() if BOT is None else BOT
+    BOT = run_skill() if BOT is None else BOT
     # args = constants.parse_argv(argv=sys.argv)
     statements = cli(constants.args)
     if constants.args.loglevel >= 50:
