@@ -1,8 +1,15 @@
-import os
-import sys
+from collections import Counter, OrderedDict, Mapping
+from decimal import Decimal
+import datetime
 import logging
-from pathlib import Path
-from collections import Counter
+import os
+from pytz import timezone
+import requests
+import string
+import sys
+
+import numpy as np
+import pandas as pd
 
 import nltk
 import nltk.corpus
@@ -13,68 +20,29 @@ import django.conf
 
 from qary import __version__
 
-from qary.etl.fileutils import basename, url_filename, path_filename
-
 # DO ACCESS KEY and SECRET need to be integrated into env
 env = Environment(spacy_lang=str, loglevel=int, name=str)
+# if you need consistent hashes from builtin hash function set env['PYTHONHASHSEED']
 
+HOME_DIR = os.path.expanduser('~')
 SRC_DIR = os.path.dirname(__file__)
 BASE_DIR = os.path.dirname(SRC_DIR)
 DATA_DIR = os.path.join(SRC_DIR, 'data')
 LOG_DIR = os.path.join(DATA_DIR, 'log')
+CONSTANTS_DIR = os.path.join(DATA_DIR, 'constants')
 HISTORY_PATH = os.path.join(DATA_DIR, 'history.json')
 os.makedirs(LOG_DIR, exist_ok=True)
+os.makedirs(CONSTANTS_DIR, exist_ok=True)
 # with open(HISTORY_PATH, 'a') as fap:
 #     pass
 MIDATA_DOMAINNAME = 'tan.sfo2.digitaloceanspaces.com'
 MIDATA_URL = f'https://{MIDATA_DOMAINNAME}'
 MIDATA_QA_MODEL_DIR = 'midata/public/models/qa'
 MIDATA_QA_MODEL_DIR_URL = f'{MIDATA_URL}{MIDATA_QA_MODEL_DIR}'
+ARTICLES_URL = f'{MIDATA_URL}/midata/public/corpora/wikipedia/articles_with_keywords.pkl'
 
 USE_CUDA = False
 MAX_TURNS = 10000
-EXIT_COMMANDS = set('exit quit bye goodbye cya'.split())
-
-
-# TODO: put this in etl/models.py or data/models.py for storage in Django ORM database
-LARGE_FILES = {
-    'wikipedia_titles': dict(
-        url='https://tan.sfo2.digitaloceanspaces.com/midata/public/corpora/enwiki-latest-all-titles-in-ns0.gz',
-        path=Path(DATA_DIR, 'corpora', 'wikipedia', 'enwiki-latest-all-titles-in-ns0.gz')),
-    'squad_dev': dict(
-        url='https://rajpurkar.github.io/SQuAD-explorer/dataset/dev-v2.0.json',
-        path=Path(DATA_DIR, 'training_sets', 'squad', 'dev-v2.0.json')),
-    'squad': dict(
-        url='https://rajpurkar.github.io/SQuAD-explorer/dataset/train-v2.0.json',
-        path=Path(DATA_DIR, 'training_sets', 'squad', 'train-v2.0.json')),
-    'floyd': dict(
-        url='https://tan.sfo2.digitaloceanspaces.com/midata/public/corpora/wikipedia/floyd.pkl',
-        path=Path(DATA_DIR, 'corpora', 'wikipedia', 'floyd.pkl')),
-    'wikipedia_articles': dict(
-        url='https://tan.sfo2.digitaloceanspaces.com/midata/public/corpora/wikipedia/articles_with_keywords.pkl',
-        path=Path(DATA_DIR, 'corpora', 'wikipedia', 'articles_with_keywords.pkl')),
-    'albert-large-v2': dict(
-        url='https://tan.sfo2.digitaloceanspaces.com/midata/public/models/qa/albert-large-v2-0.2.0.zip',
-        path=Path(DATA_DIR, 'models', 'qa', 'albert-large-v2-0.2.0.zip')),
-    'bert-base-cased': dict(
-        url='https://tan.sfo2.digitaloceanspaces.com/midata/public/models/qa/bert-base-cased-0.2.0.zip',
-        path=Path(DATA_DIR, 'models', 'qa', 'bert-base-cased-0.2.0.zip')),
-}
-tmp_large_files = {}
-for name, meta in LARGE_FILES.items():
-    m = dict()
-    m.update(meta)
-    # add redundant keys for the url and the filenames in the url
-    m['url_filename'] = url_filename(m['url'])
-    m['filename'] = path_filename(m['path']) or m['url_filename']
-    for k in m['url'], m['filename'], m['url_filename'], basename(m['url_filename']):
-        tmp_large_files[k] = m
-LARGE_FILES.update(tmp_large_files)
-del tmp_large_files, m, k
-LARGE_FILES['wikipedia'] = LARGE_FILES['wikipedia_articles']
-LARGE_FILES['albert'] = LARGE_FILES['albert-large-v2']
-LARGE_FILES['albert_large'] = LARGE_FILES['albert-large-v2']
-LARGE_FILES['albert_large_v2'] = LARGE_FILES['albert-large-v2']
 
 
 DEFAULT_CONFIG = {
@@ -83,7 +51,7 @@ DEFAULT_CONFIG = {
     'bots': 'glossary',  # glossary,qa,parul,eliza,search_fuzzy'
     'spacy_lang': 'en_core_web_sm',
     'use_cuda': USE_CUDA,
-    'loglevel': logging.FATAL,
+    'loglevel': 0,  # 0 allows the user to set a value 10-50 (lower more verbose)
     'num_top_replies': 10,
     'self_score': '.5',
     'semantic_score': '.5',
@@ -94,8 +62,8 @@ DEFAULT_CONFIG = {
 }
 DEFAULT_CONFIG.update(env.parsed)
 
-
-LOGLEVELS = [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.FATAL]
+LOGLEVELS = [
+    logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.FATAL]
 # LOG_LEVELS = [         10,           20,              30,            40,            50]
 LOGLEVEL_NAMES = 'DEBUG INFO WARNING ERROR FATAL'.split()
 LOGLEVEL_ABBREVIATIONS = [s[:4].lower() for s in LOGLEVEL_NAMES]
@@ -111,6 +79,40 @@ logging.basicConfig(
 root_logger = logging.getLogger()
 log = logging.getLogger(__name__)
 
+# TZ constants
+DEFAULT_TZ = timezone('UTC')
+
+MAX_LEN_FILEPATH = 1023  # on OSX `open(fn)` raises OSError('Filename too long') if len(fn)>=1024
+
+ROUNDABLE_NUMERIC_TYPES = (float, int, Decimal, bool)
+FLOATABLE_NUMERIC_TYPES = (float, int, Decimal, bool)
+BASIC_NUMERIC_TYPES = (float, int)
+NUMERIC_TYPES = (float, int, Decimal, complex, str)  # datetime.datetime, datetime.date
+NUMBERS_AND_DATETIMES = (float, int, Decimal, complex, str)
+SCALAR_TYPES = (float, int, Decimal, bool, complex, str)  # datetime.datetime, datetime.date
+# numpy types are derived from these so no need to include numpy.float64, numpy.int64 etc
+DICTABLE_TYPES = (Mapping, tuple, list)  # convertable to a dictionary (inherits Mapping or is a list of key/value pairs)
+VECTOR_TYPES = (list, tuple)
+PUNC = str(string.punctuation)
+
+# synonyms for "count"
+COUNT_NAMES = ['count', 'cnt', 'number', 'num', '#', 'frequency', 'probability', 'prob', 'occurences']
+# 4 types of
+
+###########
+# secrets
+SPACES_ACCESS_KEY = os.environ.get('SPACES_ACCESS_KEY')
+SPACES_SECRET_KEY = os.environ.get('SPACES_SECRET_KEY')
+
+ES_HOST = os.environ.get('ES_HOST')
+ES_PORT = os.environ.get('ES_PORT', 9200)
+ES_USER = os.environ.get('ES_USER', os.environ.get('KIBANA_USER', 'kibadmin'))
+ES_PASS = os.environ.get('ES_PASS', os.environ.get('KIBANA_PASS', 'YoUrPaSsWoRd'))
+ES_INDEX = os.environ.get('ES_INDEX', 'wikipedia')
+
+# secrets
+############
+
 
 def parse_args(args):
     """Parse command line parameters using qary.ini for the default values
@@ -121,7 +123,7 @@ def parse_args(args):
     Returns:
       :obj:`argparse.Namespace`: command line parameters namespace
     """
-    print(f"args: {args}")
+
     parser = configargparse.ArgParser(
         default_config_files=[
             '~/qary.ini',
@@ -181,6 +183,14 @@ def parse_args(args):
         help="Comma-separated list of bot personalities to load. Defaults: pattern,parul,search_fuzzy,time,eliza",
         type=str,
         metavar="STR")
+    parser.add_argument(
+        '-s',
+        '--skills',
+        default=None,  # DEFAULT_CONFIG['bots'],  # None so config.ini can populate defaults
+        dest="bots",
+        help="Comma-separated list of bot personalities to load. Defaults: pattern,parul,search_fuzzy,time,eliza",
+        type=str,
+        metavar="STR")
     parser.add_argument(
         '-q',
         '--quiet',
@@ -215,7 +225,7 @@ def parse_args(args):
         dest="loglevel",
         help="Raw integer loglevel (10=debug, 20=info, 30=warn, 40=error, 50=fatal)",
         type=int,
-        default=DEFAULT_CONFIG['loglevel'])
+        default=0)  # DEFAULT_CONFIG['loglevel'])
     parser.add_argument(
         '--spacy_lang',
         default=None,  # None allows ini to set default
@@ -230,14 +240,6 @@ def parse_args(args):
         help='Maximum n-gram length (in tokens) for wikipedia article title guesses.',
         type=int,
         metavar="INT")
-    parser.add_argument(
-        '-s',
-        '--score_weights',
-        default=DEFAULT_CONFIG['score_weights'],
-        dest="score_weights",
-        help='Dictionary of weights: {"spell": .5, "sentiment": .5, "semantic": .5}',
-        type=str,
-        metavar="DICT_STR")
     parser.add_argument(
         '--semantic',
         type=float,
@@ -272,7 +274,7 @@ def parse_args(args):
         type=str,
         metavar='STR')
     parsed_args = parser.parse_args(args)
-    print(parsed_args)
+    # print(parsed_args)
     return parsed_args
 
 
@@ -304,7 +306,7 @@ def parse_argv(argv=sys.argv):
         if command_line_path.endswith('qary') or command_line_path.endswith('bot'):
             new_argv.extend(list(argv[1:]))
     else:
-        log.error(f"It doesn't look like you're running this as a command line application. sys.argv: {argv}")
+        log.info(f"It doesn't look like you're running this as a command line application. sys.argv: {argv}")
     args = parse_args(args=new_argv)
 
     # consolidate the 2 synonymous args, loglevel and verbosity, by using the minimum of the 2
@@ -312,18 +314,16 @@ def parse_argv(argv=sys.argv):
     #   verbosity may only be set by the command line args -v/verbose -qq -vv and -q/quiet
     #   The more verbose of the 2 (lower loglevel value) wins
 
-    loglevel = min(args.loglevel or logging.WARNING, args.verbosity or logging.WARNING)
+    loglevel = min(args.loglevel or 0, args.verbosity or 0) or logging.WARNING
     setup_logging(loglevel=loglevel)
     log.debug(f'RAW ARGS (including config file): {vars(args)}')
     args.loglevel = loglevel
 
     # strip quotes in case ini file incorrectly uses single quotes that become part of the str
     args.nickname = str(args.nickname).strip().strip('"').strip("'")
-    args.bots = 'glossary,qa' if getattr(args, 'bots', None) is None else args.bots
+    args.bots = 'glossary,qa' if getattr(args, 'bots', getattr(args, 'skills', None)) is None else args.bots
     args.bots = [m.strip() for m in args.bots.split(',')]
-    log.info(f"Building a BOT with: {args.bots}")
-    log.info(f"Weights: {args.score_weights}")
-    # log.info(f"Parsed Weights: {type(json.loads(args.score_weights))}")
+    log.debug(f"Building a BOT with: {args.bots}")
 
     USE_CUDA = args.use_cuda
 
@@ -331,12 +331,14 @@ def parse_argv(argv=sys.argv):
 
 
 try:
-    # This will fail if another application (like gunicorn) imports qary and redirects stdin without running it as a command line app
-    print(f"sys.argv: {sys.argv}")
+    # This will fail if another application (like gunicorn) imports qary and redirects stdin without
+    # running it as a command line app
+    print()  # can't log anything until we know what the user's log level is.
+    # print(f"sys.argv: {sys.argv}")
     args = parse_argv(argv=sys.argv)
 except Exception as e:  # noqa
-    log.error(e)
-    log.error('Unable to parse command line arguments. Are you trying to import this into gunicorn?')
+    log.info(e)
+    log.info('Unable to parse command line arguments. Are you trying to import this into gunicorn?')
     # Workaround for the bug when Django app tries to import qary.constants:
     # `usage: gunicorn [-h] [-c CONFIG] ... gunicorn: error: unrecognized arguments
     args = parse_argv(argv=[])
@@ -434,3 +436,140 @@ def set_django_settings(settings=None):
 
 
 django_settings = set_django_settings()
+
+
+#####################################################################################
+# pugnlp.constants
+
+tld_iana = pd.read_csv(os.path.join(DATA_DIR, 'constants', 'tlds-from-iana.csv'), encoding='utf8')
+tld_iana = OrderedDict(sorted(zip((tld.strip().lstrip('.') for tld in tld_iana.domain),
+                                  [(sponsor.strip(), -1) for sponsor in tld_iana.sponsor]),
+                              key=lambda x: len(x[0]),
+                              reverse=True))
+# top 20 in Google searches per day
+# sorted by longest first so .com matches before .om (Oman)
+tld_popular = OrderedDict(sorted([
+    ('com', ('Commercial', 4860000000)),
+    ('org', ('Noncommercial', 1950000000)),
+    ('edu', ('US accredited postsecondary institutions', 1550000000)),
+    ('gov', ('United States Government', 1060000000)),
+    ('uk', ('United Kingdom', 473000000)),  # noqa
+    ('net', ('Network services', 206000000)),
+    ('ca', ('Canada', 165000000)),  # noqa
+    ('de', ('Germany', 145000000)),  # noqa
+    ('jp', ('Japan', 139000000)),  # noqa
+    ('fr', ('France', 96700000)),  # noqa
+    ('au', ('Australia', 91000000)),  # noqa
+    ('us', ('United States', 68300000)),  # noqa
+    ('ru', ('Russian Federation', 67900000)),  # noqa
+    ('ch', ('Switzerland', 62100000)),  # noqa
+    ('it', ('Italy', 55200000)),  # noqa
+    ('nl', ('Netherlands', 45700000)),  # noqa
+    ('se', ('Sweden', 39000000)),  # noqa
+    ('no', ('Norway', 32300000)),  # noqa
+    ('es', ('Spain', 31000000)),  # noqa
+    ('mil', ('US Military', 28400000)),
+    ], key=lambda x: len(x[0]), reverse=True))
+
+uri_schemes_iana = sorted(pd.read_csv(os.path.join(DATA_DIR, 'constants', 'uri-schemes.xhtml.csv'),
+                                      index_col=0).index.values,
+                          key=lambda x: len(str(x)), reverse=True)
+uri_schemes_popular = ['chrome-extension', 'example', 'content', 'bitcoin',
+                       'telnet', 'mailto',
+                       'https', 'gtalk',
+                       'http', 'smtp', 'feed',
+                       'udp', 'ftp', 'ssh', 'git', 'apt', 'svn', 'cvs']
+
+# these may not all be the sames isinstance types, depending on the env
+FLOAT_TYPES = tuple([t for t in set(np.typeDict.values()) if t.__name__.startswith('float')] + [float])
+FLOAT_DTYPES = tuple(set(np.dtype(typ) for typ in FLOAT_TYPES))
+INT_TYPES = tuple([t for t in set(np.typeDict.values()) if t.__name__.startswith('int')] + [int])
+INT_DTYPES = tuple(set(np.dtype(typ) for typ in INT_TYPES))
+NUMERIC_TYPES = tuple(set(list(FLOAT_TYPES) + list(INT_TYPES)))
+NUMERIC_DTYPES = tuple(set(np.dtype(typ) for typ in NUMERIC_TYPES))
+
+DATETIME_TYPES = [t for t in set(np.typeDict.values()) if t.__name__.startswith('datetime')]
+DATETIME_TYPES.extend([datetime.datetime, pd.Timestamp])
+DATETIME_TYPES = tuple(DATETIME_TYPES)
+
+DATE_TYPES = (datetime.datetime, datetime.date)
+
+# matrices can be column or row vectors if they have a single col/row
+VECTOR_TYPES = (list, tuple, np.matrix, np.ndarray)
+MAPPING_TYPES = (Mapping, pd.Series, pd.DataFrame)
+
+# These are the valid dates for all 3 datetime types in python (and the underelying integer nanoseconds)
+INT_MAX = INT64_MAX = 2 ** 63 - 1
+INT_MIN = INT64_MIN = - 2 ** 63
+UINT_MAX = UINT64_MAX = - 2 ** 64 - 1
+
+INT32_MAX = 2 ** 31 - 1
+INT32_MIN = - 2 ** 31
+UINT32_MAX = - 2 ** 32 - 1
+
+INT16_MAX = 2 ** 15 - 1
+INT16_MIN = - 2 ** 15
+UINT16_MAX = - 2 ** 16 - 1
+
+# Pandas timestamps can handle nanoseconds? but python datetimestamps cannot.
+MAX_TIMESTAMP = pd.Timestamp('2262-04-11 23:47:16.854775', tz='utc')
+MIN_TIMESTAMP = pd.Timestamp(datetime.datetime(1677, 9, 22, 0, 12, 44), tz='utc')
+ZERO_TIMESTAMP = pd.Timestamp('1970-01-01 00:00:00', tz='utc')
+
+# to_pydatetime() rounds to microseconds, ignoring 807 nanoseconds available in other MAX TIMESTAMPs
+MIN_DATETIME = MIN_TIMESTAMP.to_pydatetime()
+MAX_DATETIME = MAX_TIMESTAMP.to_pydatetime()
+MIN_DATETIME64 = MIN_TIMESTAMP.to_datetime64()
+MAX_DATETIME64 = MAX_TIMESTAMP.to_datetime64()
+INF = np.inf
+NAN = np.nan
+NAT = pd.NaT
+
+
+# str constants
+MAX_CHR = MAX_CHAR = chr(127)
+APOSTROPHE_CHARS = "'`’"
+# Monkey patch so import from constants if you want this:
+string.unprintable = '\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0e\x0f' \
+    '\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x7f'
+ASCII_UNPRINTABLE_CHRS = string.unprintable  # ''.join(chr(i) for i in range(128) if chr(i) not in string.printable)
+
+NULL_VALUES = set(['0', 'None', 'null', "'", ""] + ['0.' + z for z in ['0' * i for i in range(10)]])
+# if datetime's are 'repr'ed before being checked for null values sometime 1899-12-30 will come up
+NULL_REPR_VALUES = set(['datetime.datetime(1899, 12, 30)'])
+# to allow NULL checks to strip off hour/min/sec from string repr when checking for equality
+MAX_NULL_REPR_LEN = max(len(s) for s in NULL_REPR_VALUES)
+
+PERCENT_SYMBOLS = ('percent', 'pct', 'pcnt', 'pt', r'%')
+FINANCIAL_WHITESPACE = ('Flat', 'flat', ' ', ',', '"', "'", '\t', '\n', '\r', '$')
+FINANCIAL_MAPPING = (('k', '000'), ('M', '000000'))
+
+# pugnlp.constants
+#####################################################################################
+
+
+def check_es_url(es_host=ES_HOST, es_port=ES_PORT, es_user=ES_USER, es_pass=ES_PASS, es_index=ES_INDEX, es_url=None):
+    es_url = es_url or f'http://{ES_USER}:{ES_PASS}@{ES_HOST}:{ES_PORT}'
+    # TODO, use regex to check to make sure es_url contains an es_host too
+    if es_url:
+        es_index_url = f'{es_url}/{es_index}' if es_index else es_url
+        try:
+            response = requests.get(es_index_url, timeout=3)
+            log.debug(f'ElasticSearch Server Replied: {response.json()}')
+        except requests.ConnectTimeout:
+            log.info(f"ConnectTimeout while trying to connect to {es_index_url}...")
+            return None
+        except requests.ConnectionError:
+            log.info(f"Unable to connect to {es_index_url}...")
+            return None
+        except requests.Timeout:
+            log.info(f"Timeout while trying to connect to {es_index_url}...")
+            return None
+        except requests.exceptions.InvalidURL:
+            log.info(f"Invalid URL: {es_index_url}...")
+            return None
+    return es_url
+
+
+ES_URL = check_es_url()
+ES_HOST = None if ES_URL is None else ES_HOST
