@@ -1,5 +1,7 @@
 #!/usr/bin/env bash
 # release.sh
+# USAGE: source scripts/release.sh 0.5.21 'spell checker...'
+
 
 # set -e
 # pip install -U twine wheel setuptools
@@ -7,6 +9,9 @@ git commit -am "$2"
 git push
 git tag -l | cat
 git tag -a "$1" -m "$2"
+
+find qary/data -type f -size +5M -exec rm -f {} \;
+
 python setup.py sdist
 python setup.py bdist_wheel
 
@@ -17,5 +22,6 @@ if [ -z "$(which twine)" ] ; then
 fi
 
 twine check dist/*
-twine upload dist/"qary-$1"* --verbose
+twine upload dist/"qary-$1-py"* --verbose
+twine upload dist/"qary-$1.tar"* --verbose
 git push --tag
